from more_itertools import intersperse

from ..reroute import (
    StoreOnly,
    route_in,
    route_out,
)


def test_route_around():
    """Test routing of data around a consumer"""

    # Route 0 around `lambda x: 2.0 / x.
    store = []
    r = route_in(
        (
            2.0 / divisor
            for divisor in route_out(
                intersperse(0, range(2, 20)),
                store,
                lambda divisor: (StoreOnly, [divisor])
                if divisor == 0
                else (divisor, None),
            )
        ),
        store,
        lambda processed_data, stored_data: processed_data  # noqa ARG005
        if processed_data is not StoreOnly
        else 'divisor is 0',
    )
    # The result should be a list in which every odd element consists of a list
    # with the elements `[n / 2.0, n]` and every even element consists of a list
    # with `[dont_process, 0]`, because the `0`s were routed around the
    # consumer, i.e. around `lambda x: x / 2.0`.
    assert list(r) == list(intersperse('divisor is 0', (2.0 / x for x in range(2, 20))))


def test_route_no_processing():
    """Test routing of data without processing"""

    store = []
    r = route_in(
        (x for x in route_out(range(10), store, lambda x: (StoreOnly, x))),
        store,
        lambda x, y: y,  # noqa ARG005
    )
    assert list(r) == list(range(10))



"""Function to ensure that a pattern is completely contained in single chunks"""

from __future__ import annotations

import re
from typing import (
    Generator,
    Iterable,
    TypeVar,
)

S = TypeVar('S', str, bytes, bytearray)


def align_pattern(iterable: Iterable[S], pattern: S) -> Generator[S, None, None]:
    """Yield data chunks that contain a complete pattern, if it is present

    ``align_pattern`` makes it easy to find a pattern (``str``, ``bytes``,
    or ``bytearray``) in data chunks. It joins data-chunks in such a way,
    that a simple containment-check (e.g. ``pattern in chunk``) on the chunks
    that ``align_pattern`` yields will suffice to determine whether the pattern
    is present in the stream yielded by the underlying iterable or not.

    To achieve this, ``align_pattern`` will join consecutive chunks to ensures
    that the following two assertions hold:

    1. Each chunk that is yielded by ``align_pattern`` has at least the length
       of the pattern (unless the underlying iterable is exhausted before the
       length of the pattern is reached).

    2. The pattern is not split between two chunks, i.e. no chunk that is
       yielded by ``align_pattern`` ends with a prefix of the pattern (unless
       it is the last chunk that the underlying iterable yield).

    The pattern might be present multiple times in a yielded data chunk.

    Note: the ``pattern`` is compared verbatim to the content in the data
    chunks, i.e. no parsing of the ``pattern`` is performed and no regular
    expressions or wildcards are supported.

    .. code-block:: python

        >>> from datasalad.itertools import align_pattern
        >>> tuple(align_pattern([b'abcd', b'e', b'fghi'], pattern=b'def'))
        (b'abcdefghi',)
        >>> # The pattern can be present multiple times in a yielded chunk
        >>> tuple(align_pattern([b'abcd', b'e', b'fdefghi'], pattern=b'def'))
        (b'abcdefdefghi',)

    Use this function if you want to locate a pattern in an input stream. It
    allows to use a simple ``in``-check to determine whether the pattern is
    present in the yielded result chunks.

    The function always yields everything it has fetched from the underlying
    iterable. So after a yield it does not cache any data from the underlying
    iterable. That means, if the functionality of
    ``align_pattern`` is no longer required, the underlying iterator can be
    used, when ``align_pattern`` has yielded a data chunk.
    This allows more efficient  processing of the data that remains in the
    underlying iterable.

    Parameters
    ----------
    iterable: Iterable
        An iterable that yields data chunks.
    pattern: str | bytes | bytearray
        The pattern that should be contained in the chunks. Its type must be
        compatible to the type of the elements in ``iterable``.

    Yields
    -------
    bytes | bytearray
        data chunks that have at least the size of the pattern and do not end
        with a prefix of the pattern. Note that a data chunk might contain the
        pattern multiple times.
    """

    # Declare regex to be either str, bytes, or bytearray. This way we do not
    # have to discriminate between bytes and bytearray in the following
    # if-clause
    regex: str | bytes | bytearray

    # Create pattern matcher for all non-empty prefixes of the pattern
    if isinstance(pattern, str):
        regex = (
            '('
            + '|'.join(
                '.' * (len(pattern) - index - 1) + re.escape(pattern[:index]) + '$'
                for index in range(1, len(pattern))
            )
            + ')'
        )
    else:
        regex = (
            b'('
            + b'|'.join(
                b'.' * (len(pattern) - index - 1) + re.escape(pattern[:index]) + b'$'
                for index in range(1, len(pattern))
            )
            + b')'
        )
    pattern_matcher = re.compile(regex, re.DOTALL)
    pattern_sub = len(pattern) - 1
    # Join data chunks until they are sufficiently long to contain the pattern,
    # i.e. have at least size: `len(pattern)`. Continue joining, if the chunk
    # ends with a prefix of the pattern.
    current_chunk: S | None = None
    for data_chunk in iterable:
        # get the type of current_chunk from the type of this data_chunk
        if current_chunk is None:
            current_chunk = data_chunk
        else:
            current_chunk += data_chunk
        # we type-ignore the next line, because `pattern_matcher`
        # (ie. `Pattern`) only supports a subtype specification ala
        # `Pattern[str]` from Python 3.9 onwards. For now we need to
        # be compatible with Python 3.8
        if len(current_chunk) >= len(pattern) and not (
            current_chunk[-1] in pattern
            and pattern_matcher.match(current_chunk, len(current_chunk) - pattern_sub)
        ):
            yield current_chunk
            current_chunk = None

    if current_chunk is not None:
        yield current_chunk



"""Get strings decoded from chunks of bytes"""

from __future__ import annotations

from typing import (
    Generator,
    Iterable,
)

__all__ = ['decode_bytes']


def decode_bytes(
    iterable: Iterable[bytes],
    *,
    encoding: str = 'utf-8',
    backslash_replace: bool = True,
) -> Generator[str, None, None]:
    """Decode bytes in an ``iterable`` into strings

    This function decodes ``bytes`` or ``bytearray`` into ``str`` objects,
    using the specified encoding. Importantly, the decoding input can
    be spread across multiple chunks of heterogeneous sizes, for example
    output read from a process or pieces of a download.

    There is no guarantee that exactly one output chunk will be yielded for
    every input chunk. Input byte strings might be split at error-locations, or
    might be joined if a multi-byte encoding is spread over multiple chunks. If
    ``decode_bytes()`` is used together with ``itemize()``, it is advisable to
    wrap ``itemize()`` around ``decode_bytes()`` to avoid an impact on the
    number and nature of yielded items with respect to the desired itemization
    pattern.

    Multi-byte encodings that are spread over multiple byte chunks are
    supported, and chunks are joined as necessary. For example, the utf-8
    encoding for ö is ``b'\\xc3\\xb6'``.  If the encoding is split in the
    middle because a chunk ends with ``b'\\xc3'`` and the next chunk starts
    with ``b'\\xb6'``, a naive decoding approach like the following would fail:

    .. code-block:: python

       >>> [chunk.decode() for chunk in [b'\\xc3', b'\\xb6']]     # doctest: +SKIP
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "<stdin>", line 1, in <listcomp>
        UnicodeDecodeError: 'utf-8' codec can't decode byte 0xc3 in position 0: unexpected end of data

    Compared to:

    .. code-block:: python

        >>> from datasalad.itertools import decode_bytes
        >>> tuple(decode_bytes([b'\\xc3', b'\\xb6']))
        ('ö',)

    Input chunks are only joined, if it is necessary to properly decode bytes:

    .. code-block:: python

        >>> from datasalad.itertools import decode_bytes
        >>> tuple(decode_bytes([b'\\xc3', b'\\xb6', b'a']))
        ('ö', 'a')

    If ``backslash_replace`` is ``True``, undecodable bytes will be
    replaced with a backslash-substitution. Otherwise,
    undecodable bytes will raise a ``UnicodeDecodeError``:

    .. code-block:: python

        >>> tuple(decode_bytes([b'\\xc3']))
        ('\\\\xc3',)
        >>> tuple(decode_bytes([b'\\xc3'], backslash_replace=False))    # doctest: +SKIP
        Traceback (most recent call last):
            ...
        UnicodeDecodeError: 'utf-8' codec can't decode byte 0xc3 in position 1: invalid continuation byte

    Backslash-replacement of undecodable bytes is an ambiguous mapping,
    because, for example, ``b'\\xc3'`` can already be present in the input.

    Parameters
    ----------
    iterable: Iterable[bytes]
        Iterable that yields bytes that should be decoded
    encoding: str (default: ``'utf-8'``)
        Encoding to be used for decoding.
    backslash_replace: bool (default: ``True``)
        If ``True``, backslash-escapes are used for undecodable bytes. If
        ``False``, a ``UnicodeDecodeError`` is raised if a byte sequence cannot
        be decoded.

    Yields
    ------
    str
        Decoded strings that are generated by decoding the data yielded by
        ``iterable`` with the specified ``encoding``

    Raises
    ------
    UnicodeDecodeError
        If ``backslash_replace`` is ``False`` and the data yielded by
        ``iterable`` cannot be decoded with the specified ``encoding``
    """

    def handle_decoding_error(
        position: int, exc: UnicodeDecodeError
    ) -> tuple[int, str]:
        """Handle a UnicodeDecodeError"""
        if not backslash_replace:
            # Signal the error to the caller
            raise exc
        return (
            position + exc.end,
            joined_data[position : position + exc.start].decode(encoding)
            + joined_data[position + exc.start : position + exc.end].decode(
                encoding, errors='backslashreplace'
            ),
        )

    joined_data = b''
    pending_error = None
    position = 0
    for chunk in iterable:
        joined_data += chunk
        while position < len(joined_data):
            try:
                yield joined_data[position:].decode(encoding)
                joined_data = b''
                # must reset the pointer for successful decoded
                # parts too, otherwise we start too far into a new chunk's
                # content
                position = 0
            except UnicodeDecodeError as e:
                # If an encoding error occurs, we first check whether it was
                # in the middle of `joined_data` or whether it extends until the
                # end of `joined_data`.
                # If it occurred in the middle of
                # `joined_data`, we replace it with backslash encoding or
                # re-raise the decoding error.
                # If it occurred at the end of `joined_data`, we wait for the
                # next chunk, which might fix the problem.
                if position + e.end == len(joined_data):
                    # Wait for the next chunk, which might fix the problem
                    pending_error = e
                    break
                else:
                    pending_error = None
                    position, string = handle_decoding_error(position, e)
                    yield string

    if pending_error:
        # If the last chunk has a decoding error at the end, process it.
        position, string = handle_decoding_error(position, pending_error)
        if string:
            yield string



from __future__ import annotations

import pytest

from ..decode_bytes import decode_bytes


def test_split_decoding():
    encoded = 'ö'.encode()
    part_1, part_2 = encoded[:1], encoded[1:]

    # check that incomplete encodings are caught
    r = tuple(decode_bytes([b'abc' + part_1, part_2 + b'def']))
    assert ''.join(r) == 'abcödef'


def test_unfixable_error_decoding():
    encoded = 'ö'.encode()
    part_1, part_2 = encoded[:1], encoded[1:]

    # check that incomplete encodings are caught
    r = tuple(decode_bytes([b'abc' + part_1 + b'def' + part_1, part_2 + b'ghi']))
    assert ''.join(r) == 'abc\\xc3deföghi'


def test_single_undecodable_byte():
    # check that a single undecodable byte is handled properly
    r = tuple(decode_bytes([b'\xc3']))
    assert ''.join(r) == '\\xc3'
    with pytest.raises(UnicodeDecodeError):
        tuple(decode_bytes([b'\xc3'], backslash_replace=False))


def test_no_empty_strings():
    # check that empty strings are not yielded
    r = tuple(decode_bytes([b'\xc3', b'\xb6']))
    assert r == ('ö',)


def test_multiple_errors():
    r = ''.join(decode_bytes([b'08 War \xaf No \xaf More \xaf Trouble.shn.mp3']))
    assert r == '08 War \\xaf No \\xaf More \\xaf Trouble.shn.mp3'


def test_error_chunks():
    # this verifies that error handling in a previous chunk does not
    # cause data loss in a subsequent chunk
    r = ''.join(decode_bytes([b'08 War \xaf No', b'1234567890']))
    assert r == '08 War \\xaf No1234567890'



import sys

import pytest

from .. import (
    CommandError,
    iter_subproc,
)


def test_iter_subproc_cwd(tmp_path):
    test_content = 'some'
    test_file_name = 'testfile'
    test_file = tmp_path / test_file_name
    test_file.write_text(test_content)

    check_fx = (
        'import sys\n'
        f"if open('{test_file_name}').read() == '{test_content}':\n"
        "    print('okidoki')"
    )
    # we cannot read the test file without a full path, because
    # CWD is not `tmp_path`
    with pytest.raises(CommandError) as e:  # noqa PT012
        with iter_subproc([sys.executable, '-c', check_fx]):
            pass
        assert 'FileNotFoundError' in e.value

    # but if we make it change to CWD, the same code runs
    with iter_subproc([sys.executable, '-c', check_fx], cwd=tmp_path) as proc:
        out = b''.join(proc)
        assert b'okidoki' in out



import io
import sys
import threading
import time
import zipfile
from threading import Thread

import psutil
import pytest

from .iterable_subprocess import (
    CommandError,
    iterable_subprocess,
)


def test_cat_not_necessarily_streamed():
    def yield_small_input():
        yield b'first'
        yield b'second'
        yield b'third'

    with iterable_subprocess(['cat'], yield_small_input()) as output:
        assert b''.join(output) == b'firstsecondthird'


def test_cat_streamed():
    latest_input = None

    def yield_input():
        nonlocal latest_input

        for i in range(10000000):
            yield b'*' * 10
            latest_input = i

    with iterable_subprocess(['cat'], yield_input()) as output:
        latest_input_during_output = [latest_input for _ in output]

        # Make sure the input is progressing during the output. In test, there
        # are about 915 steps, so checking that it's greater than 50 shouldm't
        # make this test too flakey
        num_steps = 0
        prev_i = 0
        for i in latest_input_during_output:
            if i != prev_i:
                num_steps += 1
            prev_i = i

        assert num_steps > 50


def test_process_closed_after():
    # in datalad-next we do not necessarily have no child-processes
    # so determine the number of test incrementally
    # assert len(psutil.Process().children(recursive=True)) == 0
    n_children = len(psutil.Process().children(recursive=True))
    with iterable_subprocess(['cat'], ()) as output:
        assert len(psutil.Process().children(recursive=True)) == (n_children + 1)
    assert len(psutil.Process().children(recursive=True)) == n_children


def test_exception_from_input_before_yield_propagated():
    def yield_input():
        raise Exception('Something went wrong')

    with pytest.raises(Exception, match='Something went wrong'):
        with iterable_subprocess(['cat'], yield_input()) as output:
            pass


def test_exception_from_input_after_yield_propagated():
    def yield_input():
        yield b'*'
        raise Exception('Something went wrong')

    with pytest.raises(Exception, match='Something went wrong'):
        with iterable_subprocess(['cat'], yield_input()) as output:
            pass


def test_exception_from_input_incorrect_type_propagated():
    def yield_input():
        yield 'this-should-be-bytes'

    with pytest.raises(TypeError):
        with iterable_subprocess(['cat'], yield_input()) as output:
            pass


@pytest.mark.parametrize(
    'size',
    [
        1,
        100,
        10000,
        1000000,
    ],
)
def test_exception_from_output_during_input_iterating_propagates_and_does_not_hang(
    size,
):
    event = threading.Event()

    def yield_input():
        while True:
            event.set()
            yield b'*' * size

    with pytest.raises(Exception, match='My error'):
        with iterable_subprocess(['cat'], yield_input()) as output:
            event.wait()
            raise Exception('My error')


@pytest.mark.parametrize(
    'chunk_size',
    [
        1,
        100,
        10000,
        1000000,
    ],
)
@pytest.mark.parametrize(
    'at_iteration',
    [
        0,
        1,
        100,
    ],
)
def test_exception_from_output_iterating_propagates_and_does_not_hang(
    at_iteration, chunk_size
):
    def yield_input():
        while True:
            yield b'*' * chunk_size

    with pytest.raises(Exception, match='My error'):
        with iterable_subprocess(
            ['cat'], yield_input(), chunk_size=chunk_size
        ) as output:
            for i, chunk in enumerate(output):
                if i == at_iteration:
                    raise Exception('My error')


def test_exception_from_not_found_process_propagated():
    with pytest.raises(FileNotFoundError):
        with iterable_subprocess(['does-not-exist'], ()) as output:
            b''.join(output)


def test_exception_from_return_code(monkeypatch):
    monkeypatch.setenv('LANG', 'C')
    with pytest.raises(CommandError, match='No such file or directory') as excinfo:
        with iterable_subprocess(['ls', 'does-not-exist'], ()) as output:
            a = b''.join(output)

    assert excinfo.value.returncode > 0
    assert b'No such file or directory' in excinfo.value.stderr


def test_exception_from_context_even_though_return_code_with_long_standard_error():
    with pytest.raises(Exception, match='Another exception'):
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import sys; print("Out"); print("Error message" * 100000, file=sys.stderr); sys.exit(1)',
            ],
            (),
        ) as output:
            for _ in output:
                pass
            raise Exception('Another exception')


def test_exception_from_return_code_with_long_standard_error():
    with pytest.raises(CommandError) as excinfo:
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import sys; print("Out"); print("Error message" * 100000, file=sys.stderr); sys.exit(2)',
            ],
            (),
        ) as output:
            for _ in output:
                pass

    assert excinfo.value.returncode == 2
    assert len(excinfo.value.stderr) == 65536


def test_if_process_exits_with_non_zero_error_code_and_inner_exception_it_propagates():
    def yield_input():
        while True:
            yield b'*' * 10

    with pytest.raises(Exception, match='Another exception'):
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import sys; print("The error", file=sys.stderr); print("After output"); sys.exit(1)',
            ],
            yield_input(),
        ) as output:
            all_output = b''.join(output)
            raise Exception('Another exception')

    # rstrip to account for different platform line endings here
    assert all_output.rstrip() == b'After output'


def test_if_process_closes_standard_input_but_exits_with_non_zero_error_code_then_broken_pipe_error():
    def yield_input():
        while True:
            yield b'*' * 10

    with pytest.raises(BrokenPipeError):
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import sys; sys.stdin.close(); print("The error", file=sys.stderr); print("After output"); sys.exit(0)',
            ],
            yield_input(),
        ) as output:
            all_output = b''.join(output)

    # rstrip to account for different platform line endings here
    assert all_output.rstrip() == b'After output'


def test_if_process_closes_standard_input_but_exits_with_non_zero_error_code_then_iterable_subprocess_error():
    def yield_input():
        while True:
            yield b'*' * 10

    with pytest.raises(CommandError) as excinfo:
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import sys; sys.stdin.close(); print("The error", file=sys.stderr); print("After output"); sys.exit(3)',
            ],
            yield_input(),
        ) as output:
            all_output = b''.join(output)

    # rstrip to account for different platform line endings here
    assert all_output.rstrip() == b'After output'
    assert excinfo.value.returncode == 3
    assert excinfo.value.stderr.rstrip() == b'The error'


def test_program_that_outputs_for_a_long_time_is_interrupted_on_context_exit():
    start = time.monotonic()

    with pytest.raises(CommandError) as excinfo:
        with iterable_subprocess(
            [
                sys.executable,
                '-c',
                'import time; start = time.monotonic()\nwhile (time.monotonic() - start) < 60:\n    print("Output" * 1000)',
            ],
            (),
        ) as output:
            pass

    end = time.monotonic()

    assert excinfo.value.returncode != 0
    # alternative condition reflects error communication on windows (errno22)
    assert (
        b'BrokenPipeError' in excinfo.value.stderr
        or b'Errno 22' in excinfo.value.stderr
    )
    assert end - start < 10


def test_program_that_sleeps_exits_quickly_if_exception():
    start = time.monotonic()

    with pytest.raises(Exception, match='From context'):
        with iterable_subprocess(
            [sys.executable, '-c', 'import time; time.sleep(60)'], ()
        ) as output:
            raise Exception('From context')

    end = time.monotonic()

    assert end - start < 10


def test_program_that_sleeps_exits_quickly_if_keyboard_interrupt():
    start = time.monotonic()

    with pytest.raises(KeyboardInterrupt, match='From context'):
        with iterable_subprocess(
            [sys.executable, '-c', 'import time; time.sleep(60)'], ()
        ) as output:
            raise KeyboardInterrupt('From context')

    end = time.monotonic()

    assert end - start < 10


def test_program_that_sleeps_exits_quickly_if_keyboard_interrupt_just_before_thread_starts(
    monkeypatch,
):
    start = time.monotonic()

    def start_that_raises_keyboard_interrupt(self):
        raise KeyboardInterrupt('Just before starting thread')

    monkeypatch.setattr(Thread, 'start', start_that_raises_keyboard_interrupt)

    with pytest.raises(KeyboardInterrupt, match='Just before starting thread'):
        iterable_subprocess(
            [sys.executable, '-c', 'import time; time.sleep(60)'], ()
        ).__enter__()

    end = time.monotonic()

    assert end - start < 10


def test_program_that_sleeps_exits_quickly_if_keyboard_interrupt_just_after_thread_starts(
    monkeypatch,
):
    start = time.monotonic()

    original_start = Thread.start

    def start_that_raises_keyboard_interrupt(self):
        original_start(self)
        raise KeyboardInterrupt('Just after starting thread')

    monkeypatch.setattr(Thread, 'start', start_that_raises_keyboard_interrupt)

    with pytest.raises(KeyboardInterrupt, match='Just after starting thread'):
        iterable_subprocess(
            [sys.executable, '-c', 'import time; time.sleep(60)'], ()
        ).__enter__()

    end = time.monotonic()

    assert end - start < 10


def test_program_that_sleeps_not_quickly_if_no_exception():
    start = time.monotonic()

    with iterable_subprocess(
        [sys.executable, '-c', 'import time; time.sleep(2)'], ()
    ) as output:
        pass

    end = time.monotonic()

    assert end - start > 2


def test_funzip_no_compression():
    contents = b'*' * 100000

    def yield_input():
        file = io.BytesIO()
        with zipfile.ZipFile(file, 'w', zipfile.ZIP_STORED) as zf:
            zf.writestr('any.txt', contents)

        yield file.getvalue()

    with iterable_subprocess(['funzip'], yield_input()) as output:
        assert b''.join(output) == contents


def test_funzip_deflate():
    contents = b'*' * 100000

    def yield_input():
        file = io.BytesIO()
        with zipfile.ZipFile(file, 'w', zipfile.ZIP_DEFLATED) as zf:
            zf.writestr('any.txt', contents)

        yield file.getvalue()

    with iterable_subprocess(['funzip'], yield_input()) as output:
        assert b''.join(output) == contents


def test_error_returncode_available_from_generator():
    with pytest.raises(CommandError):
        with iterable_subprocess(['ls', 'does-not-exist'], ()) as ls:
            tuple(ls)
    assert ls.returncode != 0


def test_error_returncode_available_from_generator_with_exception():
    with pytest.raises(StopIteration):
        with iterable_subprocess(['ls', 'does-not-exist'], ()) as ls:
            while True:
                next(ls)
    assert ls.returncode != 0


def test_returncode_available_from_generator_with_exception():
    with pytest.raises(StopIteration):
        with iterable_subprocess(['echo', 'a'], ()) as echo:
            while True:
                next(echo)
    # On a Linux system, all exceptions that are raised before the subprocess
    # exited will lead to a -15 return code. On a Windows system, exceptions
    # will lead to a 1 return code. If StopIteration is raised, the
    # subprocess will either have terminated which results in a 0-return code,
    # or the subprocess is still running and will therefore be terminated which
    # results in a -15 or a 1 return code. Any other exception than
    # StopIteration, e.g. a CommandError because echo could not be found, would
    # lead to an early test-exit and not proceed to the assign-statement.
    assert echo.returncode in (0, 1, -15)



from __future__ import annotations

import json
from json.decoder import JSONDecodeError

import pytest

from ..decode_bytes import decode_bytes
from ..itemize import itemize
from ..load_json import (
    load_json,
    load_json_with_flag,
)

json_object = {
    'list1': ['a', 'bäöl', 1],
    'dict1': {
        'x': 123,
        'y': 234,
        'z': 456,
    },
}


correct_json = b'\n'.join(json.dumps(x).encode() for x in [json_object] * 10) + b'\n'

correct_chunks = [
    correct_json[i : i + 10] for i in range(0, len(correct_json) + 10, 10)
]

faulty_json = correct_json.replace(b'}\n', b'\n')
faulty_chunks = [faulty_json[i : i + 10] for i in range(0, len(correct_json) + 10, 10)]


def test_load_json_on_decoded_bytes():
    assert all(
        x == json_object
        for x in load_json(decode_bytes(itemize(correct_chunks, b'\n')))
    )
    with pytest.raises(JSONDecodeError):
        list(load_json(decode_bytes(itemize(faulty_chunks, b'\n'))))


def test_load_json_with_flag():
    assert all(
        obj == json_object and success is True
        for (obj, success) in load_json_with_flag(
            decode_bytes(itemize(correct_chunks, b'\n'))
        )
    )
    assert all(
        isinstance(exc, JSONDecodeError) and success is False
        for (exc, success) in load_json_with_flag(
            decode_bytes(itemize(faulty_chunks, b'\n'))
        )
    )
